@@ -12,17 +12,15 @@
 #	See the License for the specific language governing permissions and
 #	limitations under the License.
 
-"""Python-PouchDB_ is a Python wrapper for the `PouchDB JavaScript
-library`_. This module mostly mirrors `its API`_.
+"""The :mod:`pouchdb` package mostly mirrors the `PouchDB JavaScript
+API`_.
 
-.. _Python-PouchDB: http://python-pouchdb.marten-de-vries.nl
-.. _PouchDB JavaScript library: http://pouchdb.com/
-.. _its API: http://pouchdb.com/api.html
+.. _PouchDB JavaScript API: http://pouchdb.com/api.html
 
-There are two API's: one asynchronous one, which almost directly maps
-functions to their JavaScript equivalents and supports both promises and
-callbacks in most cases, and a synchronous one, which doesn't have a
-JavaScript equivalent, but is often easier to use.
+There are two API versions: one asynchronous one, which almost directly
+maps functions to their JavaScript equivalents and supports both
+promises and callbacks in most cases, and a synchronous one, which
+doesn't have a JavaScript equivalent, but is often easier to use.
 
 API examples
 ============
@@ -85,30 +83,6 @@ As shown in the example, you can pass in dictionaries as documents.
 Additionally, Python-PouchDB allows passing in JSON strings. So, no need
 to convert if you already have a JSON string.
 
-PouchDB Plug-ins
-================
-
-Python-PouchDB not only wraps PouchDB, but also quite a few of its
-plug-ins. The included plug-ins are as follows.
-
-**Third party plug-ins:**
-
-- `PouchDB Spatial <https://github.com/pouchdb/geopouch>`_
-- `GQL <https://github.com/pouchdb/GQL>`_ (`documentation <http://pouchdb.com/gql.html>`_)
-- `PouchDB Search <https://github.com/pouchdb/pouchdb-search>`_
-- `PouchDB Authentication <https://github.com/nolanlawson/pouchdb-authentication>`_
-
-**Plug-ins developed alongside Python-PouchDB:**
-
-- PouchDB Validation
-- PouchDB Show
-- PouchDB List
-- PouchDB Update
-
-For more information on the plug-ins developed alongside Python-PouchDB,
-see `their common webpage
-<http://python-pouchdb.marten-de-vries.nl/plugins.html>`_.
-
 """
 
 from pouchdb import context
@@ -131,6 +105,8 @@ with utils.suppress(ImportError):
 
 _IGNORED_ERRORS = [
 	"ArrayBuffer values are deprecated in Blob Constructor.",
+	"ArrayBuffer is deprecated in XMLHttpRequest.send(). Use ArrayBufferView instead.",
+	"EVALUATED FUNCTION LOGS:",
 ]
 
 def _getPromiseResponse(ctx, promise):
@@ -211,13 +187,6 @@ class AbstractPouchDB(object):
 	  database documents offline too. Unnecessary for CouchDB, but handy
 	  for e.g. PouchDB-Server, which doesn't validate itself.
 
-	Extra (optional) options for all ``AbstractPouchDB.updating*``
-	methods compared to their normal ``AbstractPouchDB.*`` counterparts
-	are:
-
-	- ``withValidation``: If ``True``, the update function uses
-	  :meth:`validatingPut` instead of (the default) :meth:`put`.
-
 	"""
 	__metaclass__ = abc.ABCMeta
 
@@ -559,7 +528,7 @@ class AbstractPouchDB(object):
 
 	validating_remove_attachment = _aliasFor("validatingRemoveAttachment")
 
-	#show plug-in - experimental so private for now
+	#show plug-in
 	@abc.abstractmethod
 	def show(self, showPath, **options):
 		"""Same as requesting ``_show`` in CouchDB. Wraps the show
@@ -567,7 +536,7 @@ class AbstractPouchDB(object):
 
 		"""
 
-	#list plug-in - experimental so private for now
+	#list plug-in
 	@abc.abstractmethod
 	def list(self, listPath, **options):
 		"""Same as requesting ``_list`` in CouchDB. Wraps the list
@@ -575,24 +544,46 @@ class AbstractPouchDB(object):
 
 		"""
 
-	#update plug-in - experimental so private for now
+	#update plug-in
 	@abc.abstractmethod
-	def updatingPut(self, url, **options):
-		"""Same as sending a PUT to ``_update`` in CouchDB. Wraps the
-		update PouchDB plug-in.
+	def update(self, url, **options):
+		"""Same as sending a request to ``_update`` in CouchDB. Wraps
+		the update PouchDB plug-in.
+
+		``options``:
+
+		- ``withValidation``: If ``True``, the update function uses
+		  :meth:`validatingPut` instead of (the default) :meth:`put`.
 
 		"""
 
-	updating_put = _aliasFor("updatingPut")
-
+	#rewrite plug-in
 	@abc.abstractmethod
-	def updatingPost(self, url, **options):
-		"""Same as sending a POST to ``_update`` in CouchDB. Wraps the
-		update PouchDB plug-in.
+	def rewrite(self, url, **options):
+		"""Same as sending a request to ``_rewrite`` in CouchDB. Wraps
+		the rewrite PouchDB plug-in. The ``url`` has the form
+		`designDocName/rewritePath`. Keep in mind that you need to use
+		``options.query`` instead of appending a ?key=value string to
+		the ``url``.
+
+		``options``:
+
+		- ``withValidation``: If ``True``, the ``db.validating*``
+		  functions are used instead of their ``db.*`` counterparts if
+		  the rewrite routes to such a resource where that is relevant.
 
 		"""
 
-	updating_post = _aliasFor("updatingPost")
+	@abc.abstractmethod
+	def rewriteResultRequestObject(self, url, **options):
+		"""See :meth:`rewrite`. Instead of returning the result of the
+		function that the rewrite routes to, this returns a CouchDB
+		request object that points to the resource the rewrite resolves
+		to. See also the CouchDB documentation `on the request object`_.
+
+		.. _on the request object: http://docs.couchdb.org/en/latest/json-structure.html#request-object
+
+		"""
 
 @contextlib.contextmanager
 def _convert404ToKeyError():
@@ -842,11 +833,15 @@ class SyncPouchDB(AbstractPouchDB, collections.MutableMapping):
 		return self._syncCall("list", listPath, options)
 
 	#update plug-in
-	def updatingPut(self, url, **options):
-		return self._syncCall("updatingPut", url, options)
+	def update(self, url, **options):
+		return self._syncCall("update", url, options)
 
-	def updatingPost(self, url, **options):
-		return self._syncCall("updatingPost", url, options)
+	#rewrite plug-in
+	def rewrite(self, url, **options):
+		return self._syncCall("rewrite", url, options)
+
+	def rewriteResultRequestObject(self, url, **options):
+		return self._syncCall("rewriteResultRequestObject", url, options)
 
 class BaseError(Exception):
 	"""The base class for all errors this module should raise.
@@ -1018,11 +1013,15 @@ class AsyncPouchDB(AbstractPouchDB):
 		return self._asyncCall("list", listPath, options, callback)
 
 	#update plug-in
-	def updatingPut(self, url, callback=None, **options):
-		return self._asyncCall("updatingPut", url, options, callback)
+	def update(self, url, callback=None, **options):
+		return self._asyncCall("update", url, options, callback)
 
-	def updatingPost(self, url, callback=None, **options):
-		return self._asyncCall("updatingPost", url, options, callback)
+	#rewrite plug-in
+	def rewrite(self, url, callback=None, **options):
+		return self._asyncCall("rewrite", url, options, callback)
+
+	def rewriteResultRequestObject(self, url, callback=None, **options):
+		return self._asyncCall("rewriteResultRequestObject", url, options, callback)
 
 class AbstractEnvironment(object):
 	"""The environment is `an event emitter
@@ -1173,9 +1172,28 @@ class AbstractEnvironment(object):
 
 	set_max_listeners = _aliasFor("setMaxListeners")
 
+	#allDbs plug-in
+	@abc.abstractmethod
+	def allDbs(self):
+		"""Wraps the pouchdb-all-dbs plug-in. Returns a list of the
+		names of all local databases.
+
+		"""
+
+	all_dbs = _aliasFor("allDbs")
+
+	@abc.abstractmethod
+	def resetAllDbs(self):
+		"""Wraps the pouchdb-all-dbs plug-in. Destroys the separate
+		allDbs database. You should never need to call this function.
+
+		"""
+
+	reset_all_dbs = _aliasFor("resetAllDbs")
+
 class AsyncEnvironment(AbstractEnvironment):
 	def __init__(self, *args, **kwargs):
-		super(AsyncEnvironment, self).__init__(AsyncPouchDB, *args, **kwargs)		
+		super(AsyncEnvironment, self).__init__(AsyncPouchDB, *args, **kwargs)
 
 	def destroy(self, name_=None, callback=None, **options):
 		if callable(name_):
@@ -1189,7 +1207,27 @@ class AsyncEnvironment(AbstractEnvironment):
 	def sync(self, source, target, **options):
 		return self._asyncCall("sync", source, target, options)
 
-class SyncEnvironment(AbstractEnvironment):
+	#allDbs plugin
+	def allDbs(self, callback=None):
+		return self._asyncCall("allDbs", callback)
+
+	def resetAllDbs(self, callback=None):
+		return self._asyncCall("resetAllDbs", callback)
+
+class SyncEnvironment(AbstractEnvironment, collections.Mapping):
+	"""This class is a Mapping, which means it can be used like a
+	:class:`dict`. (Except for setting values, in this case).
+
+	See for all the methods that being a Mapping offers, `the docs about
+	it`_.
+
+	.. _the docs about it: https://docs.python.org/2/library/collections.html#collections-abstract-base-classes
+
+	>>> env = setup()
+	>>> env.keys() #doctest: +ELLIPSIS
+	[...]
+
+	"""
 	def __init__(self, *args, **kwargs):
 		super(SyncEnvironment, self).__init__(SyncPouchDB, *args, **kwargs)		
 
@@ -1205,6 +1243,12 @@ class SyncEnvironment(AbstractEnvironment):
 		"""
 		self.destroy(dbName)
 
+	def __iter__(self):
+		return iter(self.allDbs())
+
+	def __len__(self):
+		return len(self.allDbs())
+
 	def _syncCall(self, methodName, *args):
 		promise = self._ctx.callStatic(methodName, *_filterNone(args))
 		return _getPromiseResponse(self._ctx, promise)
@@ -1231,6 +1275,13 @@ class SyncEnvironment(AbstractEnvironment):
 		"""
 		return self._callFunc(options)("sync", source, target, options)
 
+	#allDbs plug-in
+	def allDbs(self):
+		return self._syncCall("allDbs")
+
+	def resetAllDbs(self):
+		return self._syncCall("resetAllDbs")
+
 #Caching prevents identical environments from being created which saves
 #quite a bit of memory and CPU. On top of that, a reference to every
 #environment needs to be kept anyway to prevent segfaulting.
