@@ -231,10 +231,10 @@ class JSContext(object):
 		#'blob' conversion
 		with utils.suppress(TypeError, KeyError):
 			if arg["type"] == "_js_returned_blob":
-				return {
+				return utils.AttrAccessDict({
 					"type": arg["blobType"],
 					"data": base64.b64decode(arg["data"].encode("ascii")),
-				}
+				})
 		#object: recursion
 		with utils.suppress(AttributeError):
 			newArg = utils.AttrAccessDict()
