@@ -24,12 +24,15 @@ from pouchdb.tests.utils import (
 
 	BASE_URL,
 	FAST_ONLY,
+	alt_setup,
 )
 
 import pouchdb
+import pouchdb.context
 import json
 import base64
 import unittest
+import collections
 
 ONLY_TEST_VALIDATION_DOC = {
 	"_id": "_design/test",
@@ -49,7 +52,7 @@ SHOW_DOCUMENT = {
 			} else {
 				return {body: doc.description}
 			}
-		}""",
+		};\n""",
 		"args": """function (doc, req) {
 			return toJSON({args: [doc, req]});
 		}""",
@@ -58,6 +61,7 @@ SHOW_DOCUMENT = {
 				return toJSON({message: "Hello World!"});
 			});
 			provides("html", function () {
+				log({"type": "is", "html": "for", "this": "func"})
 				return "<h1>Hello World!</h1>";
 			});
 			provides("css", function () {
@@ -85,6 +89,9 @@ SHOW_DOCUMENT = {
 				return 42;
 			});
 		}""",
+		"throwingError": """function (doc, req) {
+			throw new Error("Hello World!")
+		}""",
 		"throwingErrorInProvides": """function (doc, req) {
 			provides("text", function () {
 				throw new Error("Hello World!");
@@ -143,6 +150,16 @@ UPDATE_DOCUMENT = {
 	}
 }
 
+REWRITE_DOCUMENT = {
+	"_id": "_design/test",
+	"rewrites": [
+		{
+			"from": "/test/all",
+			"to": "_list/test/ids",
+		},
+	],
+}
+
 class AlternateSignatureTests(SyncPouchDBTestCaseWithDB):
 	"""Tests alternative signatures for the internally developed
 	plug-ins that ship with Python-PouchDB. Hard to do via the Python
@@ -164,8 +181,12 @@ class AlternateSignatureTests(SyncPouchDBTestCaseWithDB):
 		promise = self._env.context.evalJs("return objects[%s].validatingPost({}, function () {});" % self._db._id)
 		self.assertPromise(promise)
 
-	def testUpdatingPost(self):
-		promise = self._env.context.evalJs("return objects[%s].updatingPost('test/test/test', function () {});" % self._db._id)
+	def testUpdate(self):
+		promise = self._env.context.evalJs("return objects[%s].update('test/test/test', function () {});" % self._db._id)
+		self.assertPromise(promise)
+
+	def testRewrite(self):
+		promise = self._env.context.evalJs("return objects[%s].rewrite('test/test/test', function () {});" % self._db._id)
 		self.assertPromise(promise)
 
 @unittest.skipIf(FAST_ONLY, "too slow")
@@ -254,15 +275,25 @@ class PluginHTTPTests(SyncPouchDBHTTPTestCase):
 
 	def testUpdate(self):
 		self._db.put(UPDATE_DOCUMENT)
-		doc, req = json.loads(self._db.updatingPut("test/args/my-id").body)
+		doc, req = json.loads(self._db.update("test/args/my-id").body)
 		self.assertIsNone(doc)
 		self.assertEqual(req["id"], "my-id")
 
+	def testRewrite(self):
+		self._db.put(REWRITE_DOCUMENT)
+
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/test/all")
+
+		self.assertEqual(cm.exception["status"], 404)
+		self.assertEqual(cm.exception["name"], "not_found")
+		self.assertEqual(cm.exception["message"], "missing_named_view")
+
 class SyncValidationSuccessTests(SyncPouchDBTestCaseWithDB):
 	"""Where the other tests just check if the wrappers work and are 
-	   barely checking behaviour, the validation tests do more. That is
-	   because the other things have JS test suites, and the validation
-	   plugin doesn't.
+		 barely checking behaviour, the validation tests do more. That is
+		 because the other things have JS test suites, and the validation
+		 plugin doesn't.
 
 	"""
 
@@ -454,7 +485,7 @@ class SyncValidationForbiddenTests(SyncPouchDBTestCaseWithDBAndDoc):
 
 	def testDesignDoc(self):
 		"""A design doc is always valid, so no matter the
-		   validate_doc_update function, the stuff below should succeed.
+			 validate_doc_update function, the stuff below should succeed.
 
 		"""
 		self._db.validatingPut({
@@ -645,6 +676,11 @@ class SyncShowTests(SyncPouchDBTestCaseWithDB):
 		self.assertEqual(result["code"], 200)
 		self.assertEqual(result["body"], "")
 
+	def testThrowingError(self):
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.show("test/throwingError")
+		self.assertEqual(cm.exception["status"], 500)
+
 	def testThrowingErrorInProvides(self):
 		with self.assertRaises(pouchdb.PouchDBError) as cm:
 			self._db.show("test/throwingErrorInProvides")
@@ -895,7 +931,7 @@ class SyncListTestsWithEmptyDesignDoc(SyncPouchDBTestCaseWithDB):
 		with self.assertRaises(pouchdb.PouchDBError) as cm:
 			self._db.list("test/test/test")
 
-		self.assertEqual(cm.exception["status"], 404)
+		self.assertEqual(cm.exception["status"], 400)
 
 class SyncListTests(SyncPouchDBTestCaseWithDB):
 	def setUp(self):
@@ -965,26 +1001,26 @@ class SyncUpdateTests(SyncPouchDBTestCaseWithDBAndDoc):
 		self._db.put(UPDATE_DOCUMENT)
 
 	def testArgs(self):
-		doc, req = json.loads(self._db.updatingPost("test/args/mytest", query={"a": 3}).body)
+		doc, req = json.loads(self._db.update("test/args/mytest", query={"a": 3}).body)
 		self.assertTrue(doc["test"])
 		self.assertEqual(req["id"], "mytest")
 		self.assertEqual(req["raw_path"], "/test/_design/test/_update/args/mytest?a=3")
 
 	def testArgsWithoutDoc(self):
-		resp = self._db.updatingPost("test/args", withValidation=True)
+		resp = self._db.update("test/args", withValidation=True)
 		doc, req = json.loads(resp.body)
 		self.assertIsNone(doc)
 		self.assertNotIn("withValidation", req)
 
 	def testUnexistingFunction(self):
 		with self.assertRaises(pouchdb.PouchDBError) as cm:
-			self._db.updatingPost("test/unexisting/mytest")
+			self._db.update("test/unexisting/mytest")
 		self.assertTrue(str(cm.exception))
 		self.assertEqual(cm.exception["name"], "not_found")
 		self.assertEqual(cm.exception["message"], "missing update function unexisting on design doc _design/test")
 
 	def testSaving(self):
-		resp = self._db.updatingPost("test/save-adding-date", body=json.dumps({
+		resp = self._db.update("test/save-adding-date", body=json.dumps({
 			"_id": "test",
 			"name": "Today"
 		}))
@@ -1009,7 +1045,7 @@ class AsyncUpdateTests(AsyncPouchDBTestCaseWithDB):
 			self.assertEqual(err["name"], "ReferenceError")
 			self.assertIn("abc", err["message"])
 
-		self._db.updatingPut("test/exception", cb)
+		self._db.update("test/exception", cb)
 		self.waitUntilCalled(cb)
 
 class AsyncUpdateWithEmptyDesignDocTests(AsyncPouchDBTestCaseWithDB):
@@ -1028,5 +1064,564 @@ class AsyncUpdateWithEmptyDesignDocTests(AsyncPouchDBTestCaseWithDB):
 			self.assertEqual(err["name"], "not_found")
 			self.assertEqual(err["message"], "missing update function unexisting on design doc _design/test")
 
-		self._db.updatingPost("test/unexisting", cb)
+		self._db.update("test/unexisting", cb)
 		self.waitUntilCalled(cb)
+
+class AsyncRewriteTests(AsyncPouchDBTestCaseWithDB):
+	def setUp(self):
+		super(AsyncRewriteTests, self).setUp()
+
+		self._db.put(REWRITE_DOCUMENT, self.cb)
+		self.waitUntilCalled(self.cb)
+
+	def testBasicUrl(self):
+		def cb(err, req):
+			self.assertEqual(req.raw_path, "/test/_design/test/_list/test/ids?k=v")
+		self._db.rewriteResultRequestObject("test/test/all", cb, query={"k": "v"})
+		self.waitUntilCalled(cb)
+
+	def testBasicResp(self):
+		def cb(err, resp):
+			self.assertEqual(err["status"], 400)
+			self.assertEqual(err["name"], "not_found")
+			self.assertIn("view named ids", err["message"])
+
+		self._db.rewrite("test/test/all", cb)
+		self.waitUntilCalled(cb)
+
+class SyncRewriteTests(SyncPouchDBTestCaseWithDB):
+	def putRewrite(self, rewrite):
+		return self._db.put({
+			"_id": "_design/test",
+			"rewrites": [rewrite],
+		})
+
+	def testHighUpPath(self):
+		self.putRewrite({
+			"from": "/highup",
+			#should be sufficiently high up.
+			"to": "../../../../../..",
+		})
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/highup")
+
+		self.assertEqual(cm.exception["status"], 404)
+		self.assertEqual(cm.exception["message"], "missing")
+
+	def testBadPath(self):
+		self.putRewrite({"from": "/badpath", "to": "../../a/b/c"})
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/badpath")
+		self.assertEqual(cm.exception["status"], 404)
+
+	def testAttachmentRewrite(self):
+		ddocResp = self.putRewrite({"from": "/attachment", "to": "/attachment"})
+		#test if put succeeds
+		resp = self._db.rewrite(
+			"test/attachment/",
+			method="PUT",
+			withValidation=True,
+			body=pouchdb.context.Blob(b"Hello World!"),
+			headers={"Content-Type": "text/plain"},
+			query={"rev": ddocResp.rev}
+		)
+		self.assertTrue(resp.ok)
+
+		#test if delete succeeds
+		resp2 = self._db.rewrite(
+			"test/attachment",
+			method="DELETE",
+			withValidation=False,
+			query={"rev": resp.rev}
+		)
+		self.assertTrue(resp2.ok)
+
+		#test if post gives a 405
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite(
+				"/test/attachment",
+				method="POST",
+				#not sure if it would be required. Playing safe here.
+				#Not that it should ever reach the rev check.
+				query={"rev": resp2.rev}
+			)
+		self.assertEqual(cm.exception["status"], 405)
+		self.assertEqual(cm.exception["name"], "method_not_allowed")
+		self.assertIn("POST", cm.exception["message"])
+
+	def testLocalDocRewrite(self):
+		self.putRewrite({"from": "/doc", "to": ".././../_local/test"})
+		resp = self._db.rewrite("test/doc", method="PUT", body='{"_id": "test"}', withValidation=True)
+		self.assertTrue(resp.ok)
+
+	def testAllDbsRewrite(self):
+		self.putRewrite({"from": "/alldbs", "to": "../../../_all_dbs"})
+		resp = self._db.rewrite("test/alldbs")
+		self.assertIsInstance(resp, collections.Sequence)
+
+		resp2 = self._db.rewriteResultRequestObject("test/alldbs")
+		self.assertEqual(resp2["path"], ["_all_dbs"])
+
+	def testPostDocRewrite(self):
+		self.putRewrite({"from": "postdoc", "to": "../../", "method": "POST"})
+
+		resp = self._db.rewrite("test/postdoc", body="{}", method="POST")
+		self.assertUuid(resp.id)
+		self.assertTrue(resp.rev.startswith("1-"))
+		self.assertTrue(resp.ok)
+
+		resp2 = self._db.rewrite("test/postdoc", body="{}", method="POST", withValidation=True)
+		self.assertUuid(resp2.id)
+		self.assertTrue(resp2.rev.startswith("1-"))
+		self.assertTrue(resp2.ok)
+
+class SyncCouchDBBasedRewriteTests(unittest.TestCase):
+	"""Based on CouchDB's rewrite test suite: rewrite.js. Not every test
+	has yet been ported, but a large amount has been.
+
+	Original test source:
+	https://github.com/apache/couchdb/blob/master/share/www/script/test/rewrite.js
+
+	"""
+	@classmethod
+	def setUpClass(cls):
+		cls._env = alt_setup()
+		cls._db = cls._env.PouchDB("test")
+
+		designDoc = {
+			"_id": "_design/test",
+			"language": "javascript",
+			"_attachments": {
+				"foo.txt": {
+					"content_type": "text/plain",
+					"data": "VGhpcyBpcyBhIGJhc2U2NCBlbmNvZGVkIHRleHQ=",
+				},
+			},
+			"rewrites": [
+				{
+					"from": "foo",
+					"to": "foo.txt"
+				},
+				{
+					"from": "foo2",
+					"to": "foo.txt",
+					"method": "GET",
+				},
+				{
+					"from": "hello/:id",
+					"to": "_update/hello/:id",
+					"method": "PUT",
+				},
+				{
+					"from": "/welcome",
+					"to": "_show/welcome",
+				},
+				{
+					"from": "/welcome/:name",
+					"to": "_show/welcome",
+					"query": {
+						"name": ":name",
+					},
+				},
+				{
+					"from": "/welcome2",
+					"to": "_show/welcome",
+					"query": {
+						"name": "user",
+					},
+				},
+				{
+					"from": "/welcome3/:name",
+					"to": "_update/welcome2/:name",
+					"method": "PUT",
+				},
+				{
+					"from": "/welcome3/:name",
+					"to": "_show/welcome2/:name",
+					"method": "GET",
+				},
+				{
+					"from": "/welcome4/*",
+					"to" : "_show/welcome3",
+					"query": {
+						"name": "*",
+					},
+				},
+				{
+					"from": "/welcome5/*",
+					"to" : "_show/*",
+					"query": {
+						"name": "*",
+					},
+				},
+				{
+					"from": "basicView",
+					"to": "_view/basicView",
+				},
+				{
+					"from": "simpleForm/basicView",
+					"to": "_list/simpleForm/basicView",
+				},
+				{
+					"from": "simpleForm/basicViewFixed",
+					"to": "_list/simpleForm/basicView",
+					"query": {
+						"startkey": 3,
+						"endkey": 8,
+					},
+				},
+				{
+					"from": "simpleForm/basicViewPath/:start/:end",
+					"to": "_list/simpleForm/basicView",
+					"query": {
+						"startkey": ":start",
+						"endkey": ":end",
+					},
+					"formats": {
+						"start": "int",
+						"end": "int",
+					},
+				},
+				{
+					"from": "simpleForm/complexView",
+					"to": "_list/simpleForm/complexView",
+					"query": {
+						"key": [1, 2],
+					},
+				},
+				{
+					"from": "simpleForm/complexView2",
+					"to": "_list/simpleForm/complexView",
+					"query": {
+						"key": ["test", {}],
+					},
+				},
+				{
+					"from": "simpleForm/complexView3",
+					"to": "_list/simpleForm/complexView",
+					"query": {
+						"key": ["test", ["test", "essai"]],
+					},
+				},
+				{
+					"from": "simpleForm/complexView4",
+					"to": "_list/simpleForm/complexView2",
+					"query": {
+						"key": {"c": 1}
+					},
+				},
+				{
+					"from": "simpleForm/complexView5/:a/:b",
+					"to": "_list/simpleForm/complexView3",
+					"query": {
+						"key": [":a", ":b"],
+					},
+				},
+				{
+					"from": "simpleForm/complexView6",
+					"to": "_list/simpleForm/complexView3",
+					"query": {
+						"key": [":a", ":b"],
+					},
+				},
+				{
+					"from": "simpleForm/complexView7/:a/:b",
+					"to": "_view/complexView3",
+					"query": {
+						"key": [":a", ":b"],
+						"include_docs": ":doc",
+					},
+					"format": {
+						"doc": "bool",
+					},
+				},
+				{
+					"from": "/",
+					"to": "_view/basicView",
+				},
+				{
+					"from": "/db/*",
+					"to": "../../*",
+				},
+			],
+			"lists": {
+				"simpleForm": r"""function(head, req) {
+					log("simpleForm");
+					send('<ul>');
+					var row, row_number = 0, prevKey, firstKey = null;
+					while (row = getRow()) {
+						row_number += 1;
+						if (!firstKey) firstKey = row.key;
+						prevKey = row.key;
+						send('\n<li>Key: '+row.key
+								 +' Value: '+row.value
+								 +' LineNo: '+row_number+'</li>');
+					}
+					return '</ul><p>FirstKey: '+ firstKey + ' LastKey: '+ prevKey+'</p>';
+				}""",
+			},
+			"shows": {
+				"welcome": """function(doc,req) {
+					return "Welcome " + req.query["name"];
+				}""",
+				"welcome2": """function(doc, req) {
+					return "Welcome " + doc.name;
+				}""",
+				"welcome3": """function(doc,req) {
+					return "Welcome " + req.query["name"];
+				}""",
+			},
+			"updates": {
+				"hello": """function(doc, req) {
+					if (!doc) {
+						if (req.id) {
+							return [{
+								_id : req.id
+							}, "New World"]
+						}
+						return [null, "Empty World"];
+					}
+					doc.world = "hello";
+					doc.edited_by = req.userCtx;
+					return [doc, "hello doc"];
+				}""",
+				"welcome2": """function(doc, req) {
+					if (!doc) {
+						if (req.id) {
+							return [{
+								_id: req.id,
+								name: req.id
+							}, "New World"]
+						}
+						return [null, "Empty World"];
+					}
+					return [doc, "hello doc"];
+				}""",
+			},
+			"views": {
+				"basicView": {
+					"map": """function(doc) {
+						if (doc.integer) {
+							emit(doc.integer, doc.string);
+						}
+						
+					}""",
+				},
+				"complexView": {
+					"map": """function(doc) {
+						if (doc.type == "complex") {
+							emit([doc.a, doc.b], doc.string);
+						}
+					}"""
+				},
+				"complexView2": {
+					"map": """function(doc) {
+						if (doc.type == "complex") {
+							emit(doc.a, doc.string);
+						}
+					}""",
+				},
+				"complexView3": {
+					"map": """function(doc) {
+						if (doc.type == "complex") {
+							emit(doc.b, doc.string);
+						}
+					}""",
+				},
+			},
+		}
+
+		def makeDocs(start, end):
+			docs = []
+			for i in range(start, end):
+				docs.append({
+					"_id": str(i),
+					"integer": i,
+					"string": str(i),
+				})
+			return docs
+
+		docs1 = makeDocs(0, 10)
+		docs2 = [
+			{"a": 1, "b": 1, "string": "doc 1", "type": "complex"},
+			{"a": 1, "b": 2, "string": "doc 2", "type": "complex"},
+			{"a": "test", "b": {}, "string": "doc 3", "type": "complex"},
+			{"a": "test", "b": ["test", "essai"], "string": "doc 4", "type": "complex"},
+			{"a": {"c": 1}, "b": "", "string": "doc 5", "type": "complex"},
+		]
+		cls._db.bulkDocs({"docs": [designDoc] + docs1 + docs2})
+
+	@classmethod
+	def tearDownClass(cls):
+		cls._db.destroy()
+
+	def testSimpleRewriting(self):
+		#GET is the default http method
+		resp = self._db.rewrite("test/foo");
+		self.assertEqual(resp.data, b"This is a base64 encoded text")
+		self.assertEqual(resp.type, "text/plain")
+
+		resp2 = self._db.rewrite("test/foo2")
+		self.assertEqual(resp2.data, b"This is a base64 encoded text")
+		self.assertEqual(resp2.type, "text/plain")
+
+	def testBasicUpdate(self):
+		#hello update world
+		doc = {"word":"plankton", "name":"Rusty"}
+		resp = self._db.post(doc)
+		self.assertTrue(resp.ok)
+		docid = resp.id
+
+		resp = self._db.rewrite("test/hello/" + docid, method="PUT")
+		self.assertEqual(resp.code, 201)
+		self.assertEqual(resp.body, "hello doc")
+		self.assertIn("charset=utf-8", resp.headers["Content-Type"])
+
+		doc = self._db.get(docid)
+		self.assertEqual(doc.world, "hello")
+
+	def testBasicShow(self):
+		resp = self._db.rewrite("test/welcome", query={"name": "user"})
+		self.assertEqual(resp.body, "Welcome user")
+
+		resp = self._db.rewrite("test/welcome/user")
+		self.assertEqual(resp.body, "Welcome user")
+
+		resp = self._db.rewrite("test/welcome2")
+		self.assertEqual(resp.body, "Welcome user")
+
+	def testWelcome3Test(self):
+		resp = self._db.rewrite("test/welcome3/test", method="PUT")
+		self.assertEqual(resp.code, 201)
+		self.assertEqual(resp.body, "New World")
+		self.assertIn("charset=utf-8", resp.headers["Content-Type"])
+
+		resp = self._db.rewrite("test/welcome3/test")
+		self.assertEqual(resp.body, "Welcome test")
+
+	def testWelcome4User(self):
+		resp = self._db.rewrite("test/welcome4/user")
+		self.assertEqual(resp.body, "Welcome user")
+
+	def testWelcome5Welcome3(self):
+		resp = self._db.rewrite("test/welcome5/welcome3")
+		self.assertEqual(resp.body, "Welcome welcome3")
+
+	def testBasicView(self):
+		resp = self._db.rewrite("test/basicView")
+		self.assertEqual(resp.total_rows, 9)
+
+	def testRootRewrite(self):
+		resp = self._db.rewrite("test/")
+		self.assertEqual(resp.total_rows, 9)
+
+	def testSimpleFormBasicView(self):
+		resp = self._db.rewrite("test/simpleForm/basicView", query={"startkey": 3, "endkey": 8})
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertNotIn("Key: 1", resp.body)
+		self.assertIn("FirstKey: 3", resp.body)
+		self.assertIn("LastKey: 8", resp.body)
+
+	def testSimpleFormBasicViewFixed(self):
+		resp = self._db.rewrite("test/simpleForm/basicViewFixed")
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertNotIn("Key: 1", resp.body)
+		self.assertIn("FirstKey: 3", resp.body)
+		self.assertIn("LastKey: 8", resp.body)
+
+	def testSimpleFormBasicViewFixedDifferentQuery(self):
+		resp = self._db.rewrite("test/simpleForm/basicViewFixed", query={"startkey": 4})
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertNotIn("Key: 1", resp.body)
+		self.assertIn("FirstKey: 3", resp.body)
+		self.assertIn("LastKey: 8", resp.body)
+
+	def testSimpleViewBasicViewPath(self):
+		resp = self._db.rewrite("test/simpleForm/basicViewPath/3/8")
+		self.assertNotIn("Key: 1", resp.body)
+		self.assertIn("FirstKey: 3", resp.body)
+		self.assertIn("LastKey: 8", resp.body)
+
+	def testSimpleFormComplexView(self):
+		resp = self._db.rewrite("test/simpleForm/complexView")
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertRegexpMatches(resp.body, "FirstKey: [1, 2]")
+
+	def testSimpleFormComplexView2(self):
+		resp = self._db.rewrite("test/simpleForm/complexView2")
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertIn("Value: doc 3", resp.body)
+
+	def testSimpleFormComplexView3(self):
+		resp = self._db.rewrite("test/simpleForm/complexView3")
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertIn("Value: doc 4", resp.body)
+
+	def testSimpleFormComplexView4(self):
+		resp = self._db.rewrite("test/simpleForm/complexView4")
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertIn("Value: doc 5", resp.body)
+
+	def testSimpleFormComplexView5WithArgs(self):
+		resp = self._db.rewrite("test/simpleForm/complexView5/test/essai")
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertIn("Value: doc 4", resp.body)
+
+	def testComplexView6WithQuery(self):
+		resp = self._db.rewrite("test/simpleForm/complexView6", query={"a": "test", "b": "essai"})
+		self.assertEqual(resp.code, 200, "with query params")
+		self.assertIn("Value: doc 4", resp.body)
+
+	def testSimpleFormComplexView7WithArgsAndQuery(self):
+		resp = self._db.rewrite("test/simpleForm/complexView7/test/essai", query={"doc": True})
+		self.assertIsInstance(resp.rows[0].doc, collections.Mapping)
+
+	def testDbWithArgs(self):
+		#The original test suite uses the 'meta' query parameter which
+		#PouchDB doesn't implement. revs_info could just be dropped in
+		#without further changes, though.
+		resp = self._db.rewrite("test/db/_design/test", query={"revs_info": True})
+		self.assertEqual(resp._id, "_design/test")
+		self.assertIsInstance(resp._revs_info, collections.Sequence)
+
+class SyncRewriteTestsWithInvalidDesignDoc(SyncPouchDBTestCaseWithDB):
+	def testEmptyDesignDoc(self):
+		self._db.put({"_id": "_design/test"})
+
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/test/all")
+
+		self.assertEqual(cm.exception["status"], 404)
+		self.assertEqual(cm.exception["name"], "rewrite_error")
+		self.assertEqual(cm.exception["message"], "Invalid path.")
+
+	def testInvalidRewrites(self):
+		self._db.put({"_id": "_design/test", "rewrites": "Hello World!"})
+
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/test/all")
+
+		self.assertEqual(cm.exception["status"], 400)
+		self.assertEqual(cm.exception["name"], "rewrite_error")
+
+	def testMissingTo(self):
+		self._db.put({"_id": "_design/test", "rewrites": [
+			{"from": "*"},
+		]})
+
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/test/all")
+
+		self.assertEqual(cm.exception["status"], 500)
+		self.assertEqual(cm.exception["name"], "error")
+		self.assertEqual(cm.exception["message"], "invalid_rewrite_target")
+
+	def testEmptyRewrites(self):
+		self._db.put({"_id": "_design/test", "rewrites": []})
+
+		with self.assertRaises(pouchdb.PouchDBError) as cm:
+			self._db.rewrite("test/test/all")
+
+		self.assertEqual(cm.exception["status"], 404)
+		self.assertEqual(cm.exception["name"], "not_found")
+		self.assertEqual(cm.exception["message"], "missing")
